/*
  Copyright (C) 2009 to 2012 Chris Vine

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <unistd.h>

#include <fstream>
#include <algorithm>
#include <ios>
#include <istream>

#include <cairo.h>
#include <pango/pango-layout.h>

#include "mainwindow.h"
#include "mount_entries.h"
#include "preferences.h"
#include "dialogs.h"
#if !(GTK_CHECK_VERSION(2,99,0))
#include "utils/cairo_handle.h"
#endif

#include <c++-gtk-utils/callback.h>
#include <c++-gtk-utils/gobj_handle.h>
#include <c++-gtk-utils/mem_fun.h>

#ifdef ENABLE_NLS
#include <libintl.h>
#endif

#ifdef HAVE_X11_XLIB_H
#include <gdk/gdkx.h>
#endif

#define MAINWIN_DRAWING_AREA_WIDTH 8
#define MAINWIN_DRAWING_AREA_HEIGHT 16


void MainWindowCB::mainwin_tool_button_clicked(GtkToolButton* button_p, void* data) {
  MainWindow* instance_p = static_cast<MainWindow*>(data);
  
  if (button_p == instance_p->prefs_button_p) instance_p->prefs_impl();
  else if (button_p == instance_p->mount_entries_button_p) instance_p->mount_entries_impl();
  else if (button_p == instance_p->about_button_p) instance_p->about_impl();
  else if (button_p == instance_p->quit_button_p) instance_p->close();
  else write_error("Callback error in MainWindowCB::mainwin_tool_button_clicked()\n", false);
}


void MainWindowCB::mainwin_mount_button_clicked(GtkWidget*, void* data) {
  static_cast<Callback::Callback*>(data)->dispatch();
}

void MainWindowCB::mainwin_noentries_button_clicked(GtkWidget*, void* data) {
  static_cast<MainWindow*>(data)->mount_entries_impl();
}

gboolean MainWindowCB::mainwin_visibility_notify_event(GtkWidget*, GdkEventVisibility* event_p,
                                                       void* data) {
  MainWindow* instance_p = static_cast<MainWindow*>(data);

  if(event_p->state == GDK_VISIBILITY_FULLY_OBSCURED) instance_p->obscured = true;
  else instance_p->obscured = false;

  return false;  // carry on processing the visibility notify event
}

gboolean MainWindowCB::mainwin_window_state_event(GtkWidget*, GdkEventWindowState* event_p,
                                                  void* data) {
  MainWindow* instance_p = static_cast<MainWindow*>(data);

  if(event_p->changed_mask & GDK_WINDOW_STATE_ICONIFIED) {
    if(event_p->new_window_state & GDK_WINDOW_STATE_ICONIFIED) instance_p->minimised = true;
    else instance_p->minimised = false;
  }

  return false;  // carry on processing the window state event
}

#if GTK_CHECK_VERSION(2,99,0)

void MainWindowCB::mainwin_button_style_updated(GtkWidget*, void* data) {
  static_cast<MainWindow*>(data)->button_style_updated_impl();
}

gboolean MainWindowCB::mainwin_drawing_area_draw(GtkWidget* drawing_area_p,
						 cairo_t* cairo_p,
						 void* data) {
  return static_cast<MainWindow*>(data)->draw_mount_indicator(drawing_area_p, cairo_p);
}

#else

void MainWindowCB::mainwin_button_style_set(GtkWidget*, GtkStyle*  prev_style_p, void* data) {
  static_cast<MainWindow*>(data)->button_style_set_impl(prev_style_p);
}

gboolean MainWindowCB::mainwin_drawing_area_expose_event(GtkWidget* drawing_area_p,
							 GdkEventExpose* event_p,
							 void* data) {
  return static_cast<MainWindow*>(data)->draw_mount_indicator(drawing_area_p, event_p);
}

#endif /* GTK_CHECK_VERSION(2,99,0) */

void MainWindowCB::mainwin_delete_data(void* data, GClosure*) {
  delete static_cast<Callback::Callback*>(data);
}

// this needs to to be a separate functor class as it keeps state
// (namely, the button count for the GTK+ callback interface)
class DisplayMountpointFunc {
  int button_count;
  MainWindow* mainwin_p;
public:
  void operator()(const SharedPtr<MountItem>& item);
  DisplayMountpointFunc(MainWindow* w): button_count(0), mainwin_p(w) {}
};


MainWindow::MainWindow(void): WinBase(0, prog_config.window_icon_h),
			      button_vbox_p(0), button_extents(0, 0),
			      obscured(false), minimised(false),
			      in_dialog(false), button_style_connection(0) {

  gtk_window_set_title(get_win(), "Mount-gtk");

  gui_vbox_p = gtk_vbox_new(false, 2);
  gtk_container_add(GTK_CONTAINER(get_win()), gui_vbox_p);

  // set up the tool bar
  GtkToolbar* toolbar_p = GTK_TOOLBAR(gtk_toolbar_new());
  gtk_box_pack_start(GTK_BOX(gui_vbox_p), GTK_WIDGET(toolbar_p),
		     false, false, 0);
  gtk_toolbar_set_show_arrow(toolbar_p, false);
#if GTK_CHECK_VERSION(2,16,0)
  gtk_orientable_set_orientation(GTK_ORIENTABLE(toolbar_p), GTK_ORIENTATION_HORIZONTAL);
#else
  gtk_toolbar_set_orientation(toolbar_p, GTK_ORIENTATION_HORIZONTAL);
#endif

  // register our own button icon size for stock icons to match size of externally
  // defined icons used in this program
  gtk_icon_size_register("MOUNT_GTK_BUTTON_SIZE", 22, 22);

  prefs_button_p = GTK_TOOL_BUTTON(gtk_tool_button_new_from_stock(GTK_STOCK_PREFERENCES));
  gtk_toolbar_insert(toolbar_p, GTK_TOOL_ITEM(prefs_button_p), -1);
  gtk_tool_item_set_tooltip_text(GTK_TOOL_ITEM(prefs_button_p),
				 gettext("Preferences"));
  g_signal_connect(G_OBJECT(prefs_button_p), "clicked",
		   G_CALLBACK(MainWindowCB::mainwin_tool_button_clicked), this);

  mount_entries_button_p = GTK_TOOL_BUTTON(gtk_tool_button_new_from_stock(GTK_STOCK_EDIT));
  gtk_toolbar_insert(toolbar_p, GTK_TOOL_ITEM(mount_entries_button_p), -1);
  gtk_tool_item_set_tooltip_text(GTK_TOOL_ITEM(mount_entries_button_p),
				 gettext("Add, amend or delete table of devices to mount"));
  g_signal_connect(G_OBJECT(mount_entries_button_p), "clicked",
		   G_CALLBACK(MainWindowCB::mainwin_tool_button_clicked), this);
  gtk_tool_button_set_label(mount_entries_button_p, gettext("Edit devices"));

  about_button_p = GTK_TOOL_BUTTON(gtk_tool_button_new_from_stock(GTK_STOCK_ABOUT));
  gtk_toolbar_insert(toolbar_p, GTK_TOOL_ITEM(about_button_p), -1);
  gtk_tool_item_set_tooltip_text(GTK_TOOL_ITEM(about_button_p),
				 gettext("About mount-gtk"));
  g_signal_connect(G_OBJECT(about_button_p), "clicked",
		   G_CALLBACK(MainWindowCB::mainwin_tool_button_clicked), this);

  quit_button_p = GTK_TOOL_BUTTON(gtk_tool_button_new_from_stock(GTK_STOCK_QUIT));
  gtk_toolbar_insert(toolbar_p, GTK_TOOL_ITEM(quit_button_p), -1);
  gtk_tool_item_set_tooltip_text(GTK_TOOL_ITEM(quit_button_p),
				 gettext("Quit"));
  g_signal_connect(G_OBJECT(quit_button_p), "clicked",
		   G_CALLBACK(MainWindowCB::mainwin_tool_button_clicked), this);

  populate_mount_points();

  g_signal_connect(G_OBJECT(get_win()), "visibility_notify_event",
                   G_CALLBACK(MainWindowCB::mainwin_visibility_notify_event), this);
  g_signal_connect(G_OBJECT(get_win()), "window_state_event",
                   G_CALLBACK(MainWindowCB::mainwin_window_state_event), this);

  gtk_widget_add_events(GTK_WIDGET(get_win()), GDK_VISIBILITY_NOTIFY_MASK);

  gtk_window_set_resizable(get_win(), false);

  // button_vbox is shown in MainWindow::populate_mount_points()
  gtk_widget_show_all(GTK_WIDGET(toolbar_p));
  gtk_widget_show(gui_vbox_p);
  gtk_widget_show(GTK_WIDGET(get_win()));

  if (prog_config.tray_icon) make_tray_icon();

  mounter.monitor_cb.connect(Callback::make(*this, &MainWindow::refresh_display));
}

void MainWindow::on_delete_event(void) {
  if (prog_config.tray_icon) gtk_widget_hide(GTK_WIDGET(get_win()));
  else close();
}

void MainWindow::prefs_impl(void) {

  if (in_dialog) return;

  in_dialog = true;

  bool old_tray_icon = prog_config.tray_icon;
  bool old_tooltip = prog_config.tooltip;
  Preferences preferences(get_win());
  preferences.exec();

  if (old_tray_icon != prog_config.tray_icon) {
    if (prog_config.tray_icon) make_tray_icon();
    else tray_icon_a.reset();
  }
  if (old_tooltip != prog_config.tooltip)
    reset_mount_tooltips();

  in_dialog = false;
}

// this is called if the user elections to change preferences about
// displaying mount point tooltips
void MainWindow::reset_mount_tooltips(void) const {

  std::for_each(mount_list.begin(), mount_list.end(),
		MemFun::make(*this, &MainWindow::reset_mount_tooltips_item));

}

// called by reset_mount_tooltips() and DisplayMountpointFunc::operator()
void MainWindow::reset_mount_tooltips_item(const SharedPtr<MountItem>& item) const {
  if (prog_config.tooltip && mounter.is_mounted(item->device)) {
    gchar* str = g_strdup_printf(gettext("Mounted at: %s"),
				 mounter.get_mount_point(item->device).c_str());
    gtk_widget_set_tooltip_text(item->button_p, str);
    g_free(str);
  }
  else gtk_widget_set_has_tooltip(item->button_p, false);
}

// this is called if any mount points change
void MainWindow::refresh_display(void) const {

  std::for_each(mount_list.begin(), mount_list.end(),
		MemFun::make(*this, &MainWindow::refresh_display_item));
}

void MainWindow::refresh_display_item(const SharedPtr<MountItem>& item) const {

  GtkWidget* box_p = gtk_bin_get_child(GTK_BIN(item->button_p));
  if (!box_p) {
    write_error("Can't find child widget of button\n", false);
    return;
  }

  GtkWidget* drawing_area_p = 0;
  GtkWidget* label_p = 0;
  GList* list_p = gtk_container_get_children(GTK_CONTAINER(box_p));
  if (list_p) {
    GList* first = g_list_first(list_p);
    GList* last = g_list_last(list_p);
    if (first) drawing_area_p = GTK_WIDGET(first->data);
    if (last) label_p = GTK_WIDGET(last->data);
    g_list_free(list_p);
  }

  if (!drawing_area_p || !label_p || drawing_area_p == label_p) {
    write_error("Can't find child widgets of button hbox\n", false);
    return;
  }

  // test for change of status
  if (mounter.is_mounted(item->device)
      && !GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(drawing_area_p),
					     "is_mounted"))) {
    gtk_widget_set_sensitive(item->button_p, true);
    gtk_label_set_text(GTK_LABEL(label_p), gettext("Unmount"));
    g_object_set_data(G_OBJECT(drawing_area_p), "is_mounted",
		      GUINT_TO_POINTER(true));
    gtk_widget_queue_draw(drawing_area_p);
    if (prog_config.tooltip) {
      gchar* str = g_strdup_printf(gettext("Mounted at: %s"),
				   mounter.get_mount_point(item->device).c_str());
      gtk_widget_set_tooltip_text(item->button_p, str);
      g_free(str);
    }
  }
  else if (!mounter.is_mounted(item->device)
	   && GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(drawing_area_p),
						 "is_mounted"))){
    gtk_widget_set_sensitive(item->button_p, true);
    gtk_label_set_text(GTK_LABEL(label_p), gettext("Mount"));
    g_object_set_data(G_OBJECT(drawing_area_p), "is_mounted",
		      GUINT_TO_POINTER(false));
    gtk_widget_queue_draw(drawing_area_p);
    if (prog_config.tooltip)
      gtk_widget_set_has_tooltip(item->button_p, false);
  }
}

void MainWindow::set_button_size(const SharedPtr<MountItem>& item) const {

  gtk_widget_set_size_request(item->button_p, button_extents.first, button_extents.second);
}

void MainWindow::operation_fail(GtkWidget* widget_p) {

  // test against button_set whether the widget in question (one of the
  // mount/unmount buttons) still exists.  (It is theoretically possible that
  // a new button may have been allocated to the same address as a defunct one and
  // the user is doing a mount/unmount operation on the new button as well as the
  // old one, but in the very unlikely event of that happening it doesn't really
  // matter - the worst that could happen is that the new button is made sensitive
  // earlier than it otherwise would (ie when a mount/unmount operation on it is still
  // in course), so leading to a duplicated attempt by the user to mount or unmount
  // on the same device
  if (button_set.find(widget_p) != button_set.end()) gtk_widget_set_sensitive(widget_p, true);
}

void MainWindow::mount_entries_impl(void) {

  if (in_dialog) return;
  in_dialog = true;

  MountEntries mount_entries(get_win());
  if (mount_entries.exec()) populate_mount_points();

  in_dialog = false;
}

void MainWindow::about_impl(void) {

  // stop the non-modal About dialog being shown via the tray
  // icon if a blocking dialog is being shown, but don't cause
  // the presence of the About dialog to prevent a subsequent
  // blocking dialog from being shown
  if (in_dialog) return;

  new AboutDialog;
  // there is no memory leak - AboutDialog is self-owning
  // and will delete itself when it is closed
}

void MainWindow::button_cb(int cb_index) {

  if (cb_index >= static_cast<int>(mount_list.size())) {
    beep();
    write_error("Yikes! Index error in MainWindow::button_cb()\n", false);
  }
  else {
    gtk_widget_set_sensitive(mount_list[cb_index]->button_p, false);

    if (mounter.is_mounted(mount_list[cb_index]->device)) {
      mounter.unmount(mount_list[cb_index]->device,
                      Callback::make(*this, &MainWindow::operation_fail,
                                     mount_list[cb_index]->button_p));
    }
    else {
      mounter.mount(mount_list[cb_index]->device,
                    Callback::make(*this, &MainWindow::operation_fail,
                                   mount_list[cb_index]->button_p));
    }
  }
}

std::pair<int, int> MainWindow::get_max_button_text_extents(void) {

  int max_width = 0;
  int max_height = 0;

  if (!mount_list.empty()) {

    std::vector<std::string> text_vec;
    text_vec.push_back(gettext("Mount"));
    text_vec.push_back(gettext("Unmount"));

    std::vector<std::string>::const_iterator iter;

    int width;
    int height;

    for (iter = text_vec.begin(), width = 0, height = 0;
	 iter != text_vec.end();
	 ++iter) {

      // any button will do to get a pango layout (we use the first button in mount_list)
      GobjHandle<PangoLayout> pango_layout_h(gtk_widget_create_pango_layout(mount_list.front()->button_p,
									    iter->c_str()));
      pango_layout_get_pixel_size(pango_layout_h, &width, &height);

      if (width > max_width) {
	max_width = width;
      }
      if (height > max_height) {
	max_height = height;
      }
    }
  }
  return std::pair<int, int>(max_width, max_height);
}

void MainWindow::button_style_updated_impl(void) {

  button_extents = get_max_button_text_extents();

  if (button_extents.first) {

    // add padding
    button_extents.first += 18;  // width
    button_extents.second += 12; // height

    // add provision for the drawing area
    button_extents.first += MAINWIN_DRAWING_AREA_WIDTH;

    // have a minimum height for the drawing area whatever font is chosen
    const int min_height = MAINWIN_DRAWING_AREA_HEIGHT + 12;
    if (button_extents.second < min_height) button_extents.second = min_height;

    std::for_each(mount_list.begin(), mount_list.end(),
		  MemFun::make(*this, &MainWindow::set_button_size));
  }
}

#if !(GTK_CHECK_VERSION(2,99,0))
void MainWindow::button_style_set_impl(GtkStyle* prev_style_p) {

  GtkStyle* new_style_p = gtk_widget_get_style(GTK_WIDGET(get_win()));
  // we only need to proceed further if the font has changed
  if (!prev_style_p || !pango_font_description_equal(prev_style_p->font_desc,
						     new_style_p->font_desc)) {

    button_style_updated_impl();
  }
}
#endif // !(GTK_CHECK_VERSION(2,99,0))

#if GTK_CHECK_VERSION(2,99,0)
bool MainWindow::draw_mount_indicator(GtkWidget* drawing_area_p,
				      cairo_t* cairo_p) {

  if (gtk_widget_get_realized(drawing_area_p)) { // if we started hidden in the system tray
                                                 // the drawing area may not yet be realised
    cairo_rectangle(cairo_p, 
		    1, 1,
		    gtk_widget_get_allocated_width(drawing_area_p) - 2,
		    gtk_widget_get_allocated_height(drawing_area_p) - 2);
    if (GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(drawing_area_p),
					   "is_mounted")))
      cairo_set_source_rgb(cairo_p, 0.7, 0.0, 0.0);

    else
      cairo_set_source_rgb(cairo_p, 0.0, 0.7, 0.0);
    cairo_fill_preserve(cairo_p);
    cairo_set_source_rgb(cairo_p, 1.0, 1.0, 1.0);
    cairo_set_line_width(cairo_p, 2);
    cairo_stroke(cairo_p);
  }
  // processing stops here
  return true;
}
#else
bool MainWindow::draw_mount_indicator(GtkWidget* drawing_area_p,
				      GdkEventExpose* event_p) {

  if (GTK_WIDGET_REALIZED(drawing_area_p)) { // if we started hidden in the system tray
                                             // the drawing area may not yet be realised

    GdkDrawable* drawable_p = GDK_DRAWABLE(drawing_area_p->window);

    // provide the cairo context incorporating the destination surface
    CairoContextScopedHandle cairo_h(gdk_cairo_create(drawable_p));

    // clip any redrawing to the area of the expose event
    cairo_rectangle(cairo_h,
		    event_p->area.x, event_p->area.y,
		    event_p->area.width, event_p->area.height);
    cairo_clip(cairo_h);

    // now draw the mount indicator
    cairo_rectangle(cairo_h, 
		    1, 1,
		    (drawing_area_p->allocation.width) - 2,
		    (drawing_area_p->allocation.height) - 2);
    if (GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(drawing_area_p),
					   "is_mounted")))
      cairo_set_source_rgb(cairo_h, 0.7, 0.0, 0.0);

    else
      cairo_set_source_rgb(cairo_h, 0.0, 0.7, 0.0);
    cairo_fill_preserve(cairo_h);
    cairo_set_source_rgb(cairo_h, 1.0, 1.0, 1.0);
    cairo_set_line_width(cairo_h, 2);
    cairo_stroke(cairo_h);
  }
  // processing stops here
  return true;
}
#endif

void MainWindow::populate_mount_points(void) {

  bool rebuild = false;
  if (button_vbox_p) {  // not program start-up - rebuild interface
    if (!mount_list.empty()) {
      g_signal_handler_disconnect(mount_list.front()->button_p,
				  button_style_connection);
    }
    gtk_widget_destroy(button_vbox_p);
    mount_list.clear();
    button_set.clear();
    rebuild = true;
  }

  gchar** device_list = 0;
  gchar** label_list = 0;

  if ((device_list = g_key_file_get_string_list(prog_config.key_file_p,
						"Mounts",
						"Devices",
						0, 0))
      && (label_list = g_key_file_get_string_list(prog_config.key_file_p,
						  "Mounts",
						  "Labels",
						  0, 0))
      && *device_list) {
    
    gchar** device_tmp_p;
    gchar** label_tmp_p;
    for (device_tmp_p = device_list, label_tmp_p = label_list;
	 device_tmp_p && *device_tmp_p && label_tmp_p && *label_tmp_p;
	 ++device_tmp_p, ++label_tmp_p) {

      SharedPtr<MountItem> mount_item_s(new MountItem);
      mount_item_s->device = *device_tmp_p;
      mount_item_s->label = *label_tmp_p;

      mount_list.push_back(mount_item_s);

    }
  }
  // nothing in this function can throw - this is fine
  // g_strfreev() does nothing if the argument is NULL
  g_strfreev(device_list);
  g_strfreev(label_list);

  if(!mount_list.empty()) {
    button_vbox_p = gtk_vbox_new(true, 2);
    gtk_box_pack_start(GTK_BOX(gui_vbox_p), button_vbox_p,
		       false, false, 0);
    std::for_each(mount_list.begin(), mount_list.end(),
		  DisplayMountpointFunc(this));

    if (rebuild) { // this is a rebuild - we have previously called set_button_size() via
                   // button_style_updated_impl() when the/ window was first realised - we
                   // now need to do it for the new set of buttons

      if (button_extents.first) { // we already had buttons before for the current style,
	                          // so reuse the button extents
	std::for_each(mount_list.begin(), mount_list.end(),
		      MemFun::make(*this, &MainWindow::set_button_size));
      }
      else {       // no button extents for the current style, start from scratch.
                   // We cannot rely on just the style changed signal connection
	           // below causing button_style_updated_imp() to have been called on
	           // program start up, because we may have started without any mount
	           // points
	button_style_updated_impl();
      }
    }
    // any button will do to respond to a button style change (we use the first
    // button in mount_list).  The call to DisplayMountpointFunc() above will
    // have constructed the buttons.
#if GTK_CHECK_VERSION(2,99,0)
    button_style_connection = g_signal_connect(G_OBJECT(mount_list.front()->button_p),
					       "style_updated",
					       G_CALLBACK(MainWindowCB::mainwin_button_style_updated),
					       this);
#else
    button_style_connection = g_signal_connect(G_OBJECT(mount_list.front()->button_p),
					       "style_set",
					       G_CALLBACK(MainWindowCB::mainwin_button_style_set),
					       this);
#endif
  }
  else {
    button_vbox_p = get_noentries_widget();
    gtk_box_pack_start(GTK_BOX(gui_vbox_p), button_vbox_p,
		       false, false, 0);
  }
  gtk_widget_show_all(button_vbox_p);
}

GtkWidget* MainWindow::get_noentries_widget(void) {

  GtkWidget* button_p = gtk_button_new();
  g_signal_connect(G_OBJECT(button_p), "clicked",
		   G_CALLBACK(MainWindowCB::mainwin_noentries_button_clicked), this);

  GtkWidget* hbox_p = gtk_hbox_new(false, 2);
  gtk_container_add(GTK_CONTAINER(button_p), hbox_p);

  GtkWidget* image_p = gtk_image_new_from_stock(GTK_STOCK_EDIT, GTK_ICON_SIZE_BUTTON);
  gtk_box_pack_start(GTK_BOX(hbox_p), image_p,
		     false, false, 0);

  GtkWidget* label_p = gtk_label_new(gettext("Add device to mount"));
  gtk_box_pack_start(GTK_BOX(hbox_p), label_p,
		     true, true, 4);
  gtk_misc_set_alignment(GTK_MISC(label_p), 0, 0.5);

  return button_p;
}

void MainWindow::tray_icon_action(TrayIcon::MenuItem item) {

  switch(item) {
  case TrayIcon::prefs:
    prefs_impl();
    break;
  case TrayIcon::devices:
    mount_entries_impl();
    break;
  case TrayIcon::about:
    about_impl();
    break;
  case TrayIcon::quit:
    close();
    break;
  default:
    write_error("Invalid action passed to MainWindow::tray_icon_action()\n");
    break;
  }
}

void MainWindow::tray_icon_activate(void) {

  GtkWidget* main_widget_p = GTK_WIDGET(get_win());
  // test GTK_WIDGET_VISIBLE() as well as the 'obscured' variable - if we call
  // hide() on the window it does not count as a visibility notify event!
#if GTK_CHECK_VERSION(2,20,0)
  if (gtk_widget_get_visible(main_widget_p)
      && !obscured
      && !minimised
      && gtk_widget_get_sensitive(main_widget_p)) {
    gtk_widget_hide(main_widget_p);
  }
#else
  if (GTK_WIDGET_VISIBLE(main_widget_p)
      && !obscured
      && !minimised
      && GTK_WIDGET_SENSITIVE(main_widget_p)) {
    gtk_widget_hide(main_widget_p);
  }
#endif
  else present();
}

void MainWindow::make_tray_icon(void) {

  tray_icon_a.reset(new TrayIcon);
  tray_icon_a->set_tooltip("Mount-gtk");
  tray_icon_a->activate_cb.connect(Callback::make(*this, &MainWindow::tray_icon_activate));
  tray_icon_a->popup_cb.connect(Callback::make(*this, &MainWindow::tray_icon_action));
}

void MainWindow::present(void) {
  
  GtkWidget* main_widget_p = GTK_WIDGET(get_win());

  // make sure the window is on the current workspace
  // (gtk_window_present() should but won't do this)
#ifdef HAVE_X11_XLIB_H
# if GTK_CHECK_VERSION(2,14,0)
  gdk_x11_window_move_to_current_desktop(gtk_widget_get_window(main_widget_p));
# else
  gdk_x11_window_move_to_current_desktop(main_widget_p->window);
# endif
#endif

  // a call to gtk_window_present() now should be enough, but the calls to
  // gtk_widget_hide() and gtk_window_deiconify() will get metacity to
  // behave reasonably (shifting workspaces seems to confuse it)
  gtk_widget_hide(main_widget_p);
  gtk_window_present(get_win());
  gtk_window_deiconify(get_win());
}

void DisplayMountpointFunc::operator()(const SharedPtr<MountItem>& item) {

  GtkWidget* frame_p = gtk_frame_new(0);
  gtk_box_pack_start(GTK_BOX(mainwin_p->button_vbox_p), frame_p,
		     false, false, 0);

  GtkWidget* hbox_p = gtk_hbox_new(false, 2);
  gtk_container_add(GTK_CONTAINER(frame_p), hbox_p);

  GtkWidget* label_p = gtk_label_new(item->label.c_str());
  gtk_box_pack_start(GTK_BOX(hbox_p), label_p,
		     true, true, 4);
  gtk_misc_set_alignment(GTK_MISC(label_p), 0, 0.5);
    
  GtkWidget* button_p = gtk_button_new();
  gtk_box_pack_start(GTK_BOX(hbox_p), button_p,
		     false, false, 0);
  g_signal_connect_data(G_OBJECT(button_p), "clicked",
			G_CALLBACK(MainWindowCB::mainwin_mount_button_clicked),
			Callback::make(*mainwin_p, &MainWindow::button_cb, button_count++),
			MainWindowCB::mainwin_delete_data, GConnectFlags(0));
  item->button_p = button_p;
  mainwin_p->button_set.insert(button_p);
  mainwin_p->reset_mount_tooltips_item(item);

  // hbox for insertion in the button
  hbox_p = gtk_hbox_new(false, 2);
  gtk_container_add(GTK_CONTAINER(button_p), hbox_p);

  GtkWidget* drawing_area_p = gtk_drawing_area_new();
  gtk_box_pack_start(GTK_BOX(hbox_p), drawing_area_p,
		     false, false, 0);
  g_object_set_data(G_OBJECT(drawing_area_p), "is_mounted",
		    GUINT_TO_POINTER(mainwin_p->mounter.is_mounted(item->device)));
#if GTK_CHECK_VERSION(2,99,0)
  g_signal_connect(G_OBJECT(drawing_area_p), "draw",
		   G_CALLBACK(MainWindowCB::mainwin_drawing_area_draw),
		   this);
#else
  g_signal_connect(G_OBJECT(drawing_area_p), "expose_event",
		   G_CALLBACK(MainWindowCB::mainwin_drawing_area_expose_event),
		   mainwin_p);
#endif
  gtk_widget_set_size_request(drawing_area_p,
			      MAINWIN_DRAWING_AREA_WIDTH,
			      MAINWIN_DRAWING_AREA_HEIGHT);
  if(mainwin_p->mounter.is_mounted(item->device)) {
    label_p = gtk_label_new(gettext("Unmount"));
  }
  else {
    label_p = gtk_label_new(gettext("Mount"));
  }
  gtk_box_pack_start(GTK_BOX(hbox_p), label_p,
		     true, true, 0);
}
