/*
  Copyright (C) 2009, 2010 and 2012 Chris Vine

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <pthread.h>
#include <string.h>

#include <memory>

#include <glib.h>

#include "mounter.h"

#include <c++-gtk-utils/pipes.h>
#include <c++-gtk-utils/fdstream.h>
#include <c++-gtk-utils/thread.h>
#include <c++-gtk-utils/shared_handle.h>
#include <c++-gtk-utils/shared_ptr.h>
#include <c++-gtk-utils/gvar_handle.h>
#include <c++-gtk-utils/cgu_config.h>

#ifdef ENABLE_NLS
#include <libintl.h>
#endif

#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS 0
#endif


// used by GIO async continuations
extern "C"  {
  static void mount_gtk_async_wrapper(GObject*, GAsyncResult* res, void* data) {
#if CGU_API_VERSION>=20
    std::unique_ptr<Callback::CallbackArg<GAsyncResult*>>
      cb(static_cast<Callback::CallbackArg<GAsyncResult*>*>(data));
#else
    std::auto_ptr<Callback::CallbackArg<GAsyncResult*> >
      cb(static_cast<Callback::CallbackArg<GAsyncResult*>*>(data));
#endif
    // we have to suppress exceptions because this executes in a glib
    // main loop.  If the program were to use thread cancellation we
    // would also need a cancel-block
    try {
      cb->dispatch(res);
    }
    catch (...) {
      g_critical("mount_gtk_async_wrapper(): "
		 "exception thrown in executing continuation callback\n");
    }
  }
}

// used by GIO async continuations
struct AsyncArgs {
  Mounter::Direction direction;
  Callback::Callback* on_fail;
  Mounter* instance;
};

#if CGU_API_VERSION>=20
typedef std::unique_ptr<Callback::Callback> deleter;
#else
typedef std::auto_ptr<Callback::Callback> deleter;
#endif

void MounterCB::mounter_mounts_changed(GUnixMountMonitor*, void* data) {
  Mounter* instance_p = static_cast<Mounter*>(data);
  instance_p->fill_cache();
  instance_p->monitor_cb();
}

Mounter::Mounter(void): monitor_h(g_unix_mount_monitor_new()) {

  g_signal_connect(G_OBJECT(monitor_h.get()), "mounts_changed",
		   G_CALLBACK(MounterCB::mounter_mounts_changed), this);
  fill_cache();
}

void Mounter::fill_cache(void) {

  mounted_device_map.clear();

  GList* mount_list_p = g_list_first(g_unix_mounts_get(0));
  GList* temp_p = mount_list_p;
  while (temp_p) {
    mounted_device_map.insert(std::pair<std::string, std::string>(
	  (char*)g_unix_mount_get_device_path(static_cast<GUnixMountEntry*>(temp_p->data)),
	  (char*)g_unix_mount_get_mount_path(static_cast<GUnixMountEntry*>(temp_p->data))));
    g_unix_mount_free(static_cast<GUnixMountEntry*>(temp_p->data));
    temp_p = g_list_next(temp_p);
  }
  g_list_free(mount_list_p);
}

std::string Mounter::get_mount_point(const std::string& device) const {
  std::map<std::string, std::string>::const_iterator iter =
    mounted_device_map.find(device);
  if (iter == mounted_device_map.end()) return std::string();
  return iter->second;
}


// 'mount' can block for a considerable period when mounting such a
// file system in an attempt to cater for network perturbation.
// Therefore, to avoid blocking the GTK+ user interface,
// Mounter::raw_mount() starts a new thread to execute
// Mounter::raw_mount_thread() which will exec() 'mount'.  We observe
// the POSIX requirement for multi-threaded programs that the only
// functions which may be called in the child process between the
// fork() and the exec() are async-signal-safe ones.  We communicate
// with the UI thread with Callback::post().
void Mounter::raw_mount(const std::string& device,
			Direction direction,
			Callback::Callback* on_fail) {

  deleter del(on_fail);

  // now block off the signals with which the main thread may be interested
  // rather than this one
  sigset_t sig_mask;
  sigemptyset(&sig_mask);
  sigaddset(&sig_mask, SIGCHLD);
  sigaddset(&sig_mask, SIGQUIT);
  sigaddset(&sig_mask, SIGTERM);
  sigaddset(&sig_mask, SIGINT);
  sigaddset(&sig_mask, SIGHUP);
  pthread_sigmask(SIG_BLOCK, &sig_mask, 0);

  char* device_arg = strdup(device.c_str());

  if (!Thread::Thread::start(Callback::make(*this, &Mounter::raw_mount_thread,
					    std::pair<char*, Direction>(device_arg,
									direction),
					    del.release()),
			     false).get()) {
    free(device_arg);
    if (del.get()) Callback::post(del.release());
    write_error("Cannot start new thread to mount/unmount non-block device\n", false);
  }

  // now unblock the signals so that the initial (GUI) thread can receive them
  pthread_sigmask(SIG_UNBLOCK, &sig_mask, 0);
}

void Mounter::raw_mount_thread(std::pair<char*, Direction> mount_args,
			       Callback::Callback* on_fail) {

  ScopedHandle<char*, CFree> device_h(mount_args.first);
  deleter del(on_fail);

  PipeFifo fork_pipe(PipeFifo::block);

  // now fork to create the process which invokes 'mount'
  pid_t pid = fork();

  if (pid == -1) {
    if (del.get()) Callback::post(del.release());
    Callback::post(Callback::make(write_error,
				  "Fork error in Mounter::raw_mount_thread(\n",
				  false));
    return;
  }
  
  if (!pid) { // child process
  
    // unblock signals for mount
    sigset_t sig_mask;
    sigemptyset(&sig_mask);
    sigaddset(&sig_mask, SIGCHLD);
    sigaddset(&sig_mask, SIGQUIT);
    sigaddset(&sig_mask, SIGTERM);
    sigaddset(&sig_mask, SIGINT);
    sigaddset(&sig_mask, SIGHUP);
    // this child process is single threaded, so we can use sigprocmask()
    // rather than pthread_sigmask() (and should do so as sigprocmask()
    // is guaranteed to be async-signal-safe)
    // this process will not be receiving interrupts so we do not need
    // to test for EINTR on the call to sigprocmask()
    sigprocmask(SIG_UNBLOCK, &sig_mask, 0);

    fork_pipe.connect_to_stderr();

    if (mount_args.second == do_mount)
      execlp("mount", "mount", device_h.get(), static_cast<char*>(0));
    else
      execlp("umount", "umount", device_h.get(), static_cast<char*>(0));

    // if we reached this point, then the execlp() call must have failed
    const char message[] = "Exec failed in Mounter::raw_mount_thread().\n"
                           "Is mount installed?\n";
    write(2, message, sizeof(message) - 1);
    _exit(EXIT_SUCCESS - 1);
  }

  // read from the pipe
  fork_pipe.make_readonly();

  fdistream filein(fork_pipe.get_read_fd(), false); // fork_pipe will manage the file
                                                    // descriptor, not filein
  std::string line;
  std::string err;
  // if we use std::getline() we will be safe against
  // multi-byte encodings such as UTF-8
  while (std::getline(filein, line)) {
    err += line;
    err += '\n';
  }

  pid_t ret;
  int status;
  do {
    ret = waitpid(pid, &status, 0);
  } while (ret == -1 && errno == EINTR);
  
  bool success = false;
  if (ret != -1                                    // waitpid() did not return an error
      && WIFEXITED(status)                         // mount exited normally
      && WEXITSTATUS(status) == EXIT_SUCCESS) {    // mount returned success
    success = true;
  }
  if (!success) {
    if (!err.empty()) err += '\n';
    err += gettext("Error mounting or unmounting following device: ");
    err += device_h.get();
    err += '\n';
    if (del.get()) Callback::post(del.release());
    Callback::post(Callback::make(write_error_del,
				  const_cast<const char*>(strdup(err.c_str())),
				  true));
  }
}

void Mounter::udisks_mount(const std::string& path,
			   Direction direction,
			   Callback::Callback* on_fail) {

  // C++89 doesn't have lambdas, so use a struct in function scope
  // with the continuations as static methods
  struct AsyncChain {
    static void continuation1(const std::string& path,
			      const AsyncArgs& args) {
      deleter del(args.on_fail);
      g_dbus_proxy_new(args.instance->system_bus,
		       G_DBUS_PROXY_FLAGS_NONE,
		       0,
		       "org.freedesktop.UDisks",
		       path.c_str(),
		       "org.freedesktop.UDisks.Device",
		       0,
		       mount_gtk_async_wrapper,
		       Callback::make_ref(&AsyncChain::continuation2,
					  args));
      del.release();
    }

    static void continuation2(const AsyncArgs& args,
			      GAsyncResult* res) {
      deleter del(args.on_fail);
      GError* error = 0;
      GobjHandle<GDBusProxy> proxy(g_dbus_proxy_new_finish(res, &error));
      if (!proxy.get()) {
	del->dispatch();
	std::string err("Failed to create dbus proxy object: ");
	err += error->message;
	write_error(err.c_str(), true);
	g_error_free(error);
      }
      else {
	if (args.direction == do_mount) {
	  g_dbus_proxy_call(proxy,
			    "FilesystemMount",
			    g_variant_new("(sas)",
					  "",
					  static_cast<void*>(0)),
			    G_DBUS_CALL_FLAGS_NONE,
			    -1,
			    0,
			    mount_gtk_async_wrapper,
			    Callback::make_ref(&AsyncChain::continuation3,
					       proxy,
					       args));
	}
	else {
	  g_dbus_proxy_call(proxy,
			    "FilesystemUnmount",
			    g_variant_new("(as)",
					  static_cast<void*>(0)),
			    G_DBUS_CALL_FLAGS_NONE,
			    -1,
			    0,
			    mount_gtk_async_wrapper,
			    Callback::make_ref(&AsyncChain::continuation3,
					       proxy,
					       args));
	}
	del.release();
      }
    }

    static void continuation3(const GobjHandle<GDBusProxy>& proxy,
			      const AsyncArgs& args,
			      GAsyncResult* res) {
      deleter del(args.on_fail);
      GError* error = 0;
      GvarHandle mount_point(g_dbus_proxy_call_finish(proxy,
						      res,
						      &error));
      if (error) {
	del->dispatch();
	if (g_dbus_error_is_remote_error(error))
	  g_dbus_error_strip_remote_error(error);
	std::string err("Failed to mount or unmount device: ");
	err += error->message;
	write_error(err.c_str(), true);
	g_error_free(error);
      }
      // we are at the end of the async continuation chain, so don't
      // release del here - we need to delete on_fail
    }
  };

  // kick off the asynchronous continuation chain
  AsyncArgs args = {direction, on_fail, this};
  AsyncChain::continuation1(path, args);
}

void Mounter::get_object_path(const std::string& device,
			      Direction direction,
			      Callback::Callback* on_fail) {
  // check pre-conditions
  struct stat statbuf;
  if (stat(device.c_str(), &statbuf) || !S_ISBLK(statbuf.st_mode)) {
    raw_mount(device, direction, on_fail);
    return;
  }

  // C++89 doesn't have lambdas, so use a struct in function scope
  // with the continuations as static methods
  struct AsyncChain {
    static void continuation1(std::pair<gint64, gint64> ids,
			      const AsyncArgs& args) {
      deleter del(args.on_fail);
      g_dbus_proxy_new(args.instance->system_bus,
		       G_DBUS_PROXY_FLAGS_NONE,
		       0,
		       "org.freedesktop.UDisks",
		       "/org/freedesktop/UDisks",
		       "org.freedesktop.UDisks",
		       0,
		       mount_gtk_async_wrapper,
		       Callback::make_ref(&AsyncChain::continuation2,
					  ids,
					  args));
      del.release();
    }

    static void continuation2(std::pair<gint64, gint64> ids,
			      const AsyncArgs& args,
			      GAsyncResult* res) {
      deleter del(args.on_fail);
      GError* error = 0;
      GobjHandle<GDBusProxy> proxy(g_dbus_proxy_new_finish(res, &error));
      if (!proxy.get()) {
	del->dispatch();
	std::string err("Failed to create dbus proxy object: ");
	err += error->message;
	write_error(err.c_str(), true);
	g_error_free(error);
      }
      else {
	g_dbus_proxy_call(proxy,
			  "FindDeviceByMajorMinor",
			  g_variant_new("(xx)",
					ids.first,
					ids.second),
			  G_DBUS_CALL_FLAGS_NONE,
			  -1,
			  0,
			  mount_gtk_async_wrapper,
			  Callback::make_ref(&AsyncChain::continuation3,
					     proxy,
					     args));
	del.release();
      }
    }

    static void continuation3(const GobjHandle<GDBusProxy>& proxy,
			      const AsyncArgs& args,
			      GAsyncResult* res) {
      deleter del(args.on_fail);
      GError* error = 0;
      GvarHandle path(g_dbus_proxy_call_finish(proxy,
					       res,
					       &error));
      if (!path.get()) {
	del->dispatch();
	if (g_dbus_error_is_remote_error(error))
	  g_dbus_error_strip_remote_error(error);
	std::string err("Failed to obtain device object path: ");
	err += error->message;
	write_error(err.c_str(), true);
	g_error_free(error);
      }
      else {
	// the call to FindDeviceByMajorMinor returns a variant
	// holding a tuple holding a string.  We need to decompose
	// to get to the string
	args.instance->udisks_mount(
	       g_variant_get_string(Cgu::GvarHandle(g_variant_get_child_value(path, 0)),
				    0),
	       args.direction,
	       del.release());
      }
    }
  };
  
  // kick off the asynchronous continuation chain
  AsyncArgs args = {direction, on_fail, this};
  AsyncChain::continuation1(std::pair<gint64, gint64>((major(statbuf.st_rdev)),
						      (minor(statbuf.st_rdev))),
			    args);
}

void Mounter::mount_prepare(const std::string& device,
			    Direction direction,
			    Callback::Callback* on_fail) {
  deleter del(on_fail);

  // use lazy evaluation of the system bus connection.  The
  // synchronous version is OK here as we only call it once and
  // obtaining the system bus should be reasonably speedy
  if (!system_bus.get()) {
    GError* error = 0;
    system_bus.reset(g_bus_get_sync(G_BUS_TYPE_SYSTEM, 0, &error));
    if (!system_bus.get()) {
      del->dispatch();
      std::string err("Failed to open connection to system bus: ");
      err += error->message;
      write_error(err.c_str(), true);
      g_error_free(error);
    }
  }
  get_object_path(device, direction, del.release());
}

void Mounter::mount(const std::string& device, Callback::Callback* on_fail) {
  deleter del(on_fail);
  if (is_mounted(device)) {
    del->dispatch();
    return;
  }
  mount_prepare(device, do_mount, del.release());
}

void Mounter::unmount(const std::string& device, Callback::Callback* on_fail) {
  deleter del(on_fail);
  if (!is_mounted(device)) {
    del->dispatch();
    return;
  }
  mount_prepare(device, do_unmount, del.release());
}
